var board;      // Game Board
var intro;      // Repeating sound

var gameStatus = {
    timerEnabled: false,
    elapsedTime: 0,
    tilesMarked: 0,
    finished: false
}      

// Returns an array of all flipped tiles
var tilesFlipped = function() {
    return $$('div[data-lock="true"]').length;
}

// Generates a random number
var random = function(max) {
    return parseInt(Math.random() * max);
}

// Load the sounds and Play the intro
var initSound = function() {
    
    // Called when a sound loads
    createjs.Sound.on("fileload", function(event) {
        var url = window.location.search;
        
        // Debuting, if url contains ?nomusic don't 
        // autoplay the background music
        if (url.includes("nomusic")) {
            $('btnMusic').innerHTML = "Music ON";
        } else {
            // if the sound loaded is intro then start it
            if (event.id == "intro")
                intro = createjs.Sound.play("intro", {loop:-1, delay:0, volume:.5}); 
        }
    }, this);
    
    createjs.Sound.alternateExtensions = ["mp3"];
    createjs.Sound.registerSound("intro.ogg", "intro");
    createjs.Sound.registerSound("bomb.ogg", "bomb");
    createjs.Sound.registerSound("move.ogg", "move");
    createjs.Sound.registerSound("flip.ogg", "flip");
    createjs.Sound.registerSound("over.ogg", "over");
    createjs.Sound.registerSound("error.ogg", "error");
    createjs.Sound.registerSound("win.ogg", "win");
}

// Play a sound
var playSound = function(name) {
    createjs.Sound.play(name, {loop:0, delay:0, volume:.3}); 
}

// Update the top status
var timer = function() {
    if (gameStatus.timerEnabled) {
        gameStatus.elapsedTime += 1;
        
        // Update the timer
        var ms = gameStatus.elapsedTime * 1000;
        var secs = ms / 1000;
        var minutes = secs / 60; secs = secs % 60;
        var hours = minutes / 60; minutes = minutes % 60;
        var days = hours / 24; hours = hours % 24;
        
        var elapsed = {
            'DD': String(parseInt(days)).padStart(2, "0"),
            'HH': String(parseInt(hours)).padStart(2, "0"),
            'MM': String(parseInt(minutes)).padStart(2, "0"),
            'SS': String(parseInt(secs)).padStart(2, "0")
        }
        $('timer').innerHTML = `${elapsed.HH}h ${elapsed.MM}m ${elapsed.SS}s`;
        
        // Update the number of marks left
        var marksLeft = 10 - gameStatus.tilesMarked;
        $('marked').innerHTML = `${marksLeft} x <img src="flag.png">`;
    }
}

// Create the 2D array and build the game board
var initBoard = function() {
    board = new Array(7);
    for (var row=0; row<=7; row++) board[row] = new Array(7);
    
    // Initialize all tiles to a value of 0
    // Bombs have a value of X
    for (var row=0; row<=7; row++) 
        for (var col=0; col<=7; col++) 
            board[row][col] = 0;
    
    // ***** Board Generation: Placing Bombs

    // Keep generating random numbers till you place 
    // 10 bombs. (Don't place them on top of each other)
    var placed = 0;
    
    do {
        var row = random(7);
        var col = random(7);
        
        if (board[row][col] == 0) {
            board[row][col] = "X";
            placed+=1;
        }
    } while (placed != 10)
    
    // ***** Board Generation: Determine numbers for each tile
    
    // Helper function: Is there a bomb in this tile?
    var isBomb = function(row, col){
        // Out of bounds
        if (row < 0 || col < 0 || row > 7 || col > 7) return 0;
        
        // If this is a bomb return 1 
        return board[row][col] == 'X' ? 1 : 0;
    }
    
    // Process each tile
    for (var row=0; row<=7; row++) {
        for (var col=0; col<=7; col++) {
            
            // If this tile is a bomb, skip it
            if (board[row][col] != "X") {
                var total = 0;
                total += isBomb(row - 1, col - 1);
                total += isBomb(row - 1, col);
                total += isBomb(row - 1, col + 1);
                total += isBomb(row, col - 1);
                total += isBomb(row, col + 1);
                total += isBomb(row + 1, col - 1);
                total += isBomb(row + 1, col);
                total += isBomb(row + 1, col + 1);
                
                // Update with the number of bombs found
                board[row][col] = total;
            }
        }
    }
}

//****************** UI: Render the game board

var render = function() {
    var root = $('board');
    
    for (var r=0; r<=7; r++) {
        var row = document.createElement('div');
        for (var c=0; c<=7; c++) {
            var cell = document.createElement('div');
            
            // Border
            if (r != 0) cell.style.borderTop = "1px solid gray";
            if (c != 0) cell.style.borderLeft = "1px solid gray";
            
            // Cell Attributes / Properties
            cell.setAttribute("data-row", r);
            cell.setAttribute("data-col", c);
            cell.setAttribute("data-lock", false);
            
            cell.style.cursor = "pointer";
            cell.style.textAlign = "center";
            cell.style.verticalAlign="middle";
            cell.innerText = "";
            
            // Image
            var img = document.createElement('img');
            img.style.opacity = 0;
            img.setAttribute("data-row", r);
            img.setAttribute("data-col", c);
            cell.append(img);
            
            var mc = new Hammer(cell);
            var instance = this;

            // Left Click
            mc.on("tap press", function(ev) {
                if (ev.type == "tap") {
                    instance.reveal(ev.target.getAttribute("data-row"), ev.target.getAttribute("data-col"));
                } else if (ev.type == "press") {
                    var r = parseInt(ev.target.getAttribute("data-row"));
                    var c = parseInt(ev.target.getAttribute("data-col"));
                    
                    var obj = document.querySelector(`div[data-row="${r}"][data-col="${c}"]`);
                    
                    if (obj.style.cursor == "pointer")
                    instance.placeMarker(r, c);
                } else console.log(ev.type);
            });

            // Right Click
            cell.addEventListener("contextmenu", function(e) {
                e.preventDefault();
                var r = parseInt(this.getAttribute("data-row"));
                var c = parseInt(this.getAttribute("data-col"));
                
                var obj = document.querySelector(`div[data-row="${r}"][data-col="${c}"]`);
                
                if (obj.style.cursor == "pointer")
                placeMarker(r, c);
            });
            
            row.append(cell);
        }
        root.append(row);
    }
}

//****************** UI: Mark a tile

var placeMarker = function(r, c) {
    
    // The game is over, give it up!
    if (gameStatus.finished) return false;
    if (!gameStatus.timerEnabled) gameStatus.timerEnabled = true;
    
    // Ensure the message box is hidden and
    // lower the volume so the user can concentrate :-)
    if ($('message').style.opacity == 1) {
        hideDialogue();
        if (intro) intro.volume = .1;
    }
    
    // Get the tile the user clicked and it's image
    var obj = document.querySelector(`div[data-row="${r}"][data-col="${c}"]`);
    var img = obj.getElement("img");
    
    // Is this tile marked with a flag?
    if (img.src.includes("flag.png")) {
        
        // *** Toggle tile to having no value / reset
        
        playSound("flip"); 
        
        // Restore click event
        obj.onclick = function() {
            reveal(this.getAttribute("data-row"), this.getAttribute("data-col"));
        };
        
        // Unlock this tile
        obj.setAttribute("data-lock", false);
        gameStatus.tilesMarked -= 1;
        
        // Animation Stuff
        var i = img;
        img.velocity({ opacity: 0 },{
            duration: 100,
            queue: false,
            easing: "swing",
            complete: function() {
                i.src = "";
            },
            loop: false
        });
        
        obj.velocity({
            'background-color': "#cacaca"
        }, 100, "swing");
        
    } else {
        
        // DO you have any markers left?                
        if (gameStatus.tilesMarked >= 10) {
            playSound("error"); 
            return false;
        } 
        
        playSound("flip"); 
        
        // Deactivate click events and set the slag
        obj.onclick = "";
        img.src = "flag.png";
        
        // Lock this tile
        obj.setAttribute("data-lock", true);
        gameStatus.tilesMarked += 1;
        
        // Animation Stuff
        img.velocity({ opacity: 1 }, { 
            queue: false,
            duration: 100,
        });
        
        //'background-color': "rgb(238, 238, 238)"
        obj.velocity({
            'background-color': "rgb(167,199,220)"
        }, 100, "swing");
    }
    
    // Check for a win
    if (tilesFlipped() == 64) {
        if (intro) intro.stop();
        playSound("win"); 
        gameStatus.timerEnabled = false;
        showDialogue('You Win!<br><br><span class="yellow">Reload for a rematch!</span>'); 
    }
}

// Reveal this cell and if it's a 0 then
// reveal the cells next to it as well and if they are 0
// reveal whats next to them (recursive function)
var reveal = function(r, c) {
    
    // The game is over, give it up!
    if (gameStatus.finished) return false;
    if (!gameStatus.timerEnabled) gameStatus.timerEnabled = true;
    
    // Ensure the message box is hidden and
    // lower the volume so the user can concentrate :-)
    if ($('message').style.opacity == 1) {
        hideDialogue();
        if (intro) intro.volume = .1;
    }
    
    var row = parseInt(r);
    var col = parseInt(c);
    
    // Location out of bounds
    if (row < 0 || col < 0 || row > 7 || col > 7) {
        return false;
        
    } else {
        
        // Get the tile and its image
        var obj = document.querySelector(`div[data-row="${row}"][data-col="${col}"]`);
        var img = obj.getElement("img");
        
        // Is this tile marked or already used?
        if (obj.getAttribute("data-lock") == "true") 
        return false;
        
        // Deactivate click and prepare html control
        obj.onclick="";
        obj.style.cursor = "default";
        obj.style.fontSize = "0px";
        
        // The user clicked on a bomb
        if (board[row][col] == "X") {
            playSound("bomb"); 
            
            // Game over GI
            img.src = "bomb.png";
            gameStatus.finished = true;
            
            // Wait 2 sec and display message
            setTimeout(function() { 
                if (intro) intro.stop();
                playSound("over"); 
                gameStatus.timerEnabled = false;
                showDialogue('<div class="big">Game Over</div><br><span class="yellow">Reload to loose again!</span>', false); 
            }, 2000);
            
            // Animation Stuff
            img.velocity({ opacity: 1 }, { 
                queue: false,
                duration: 100,
            });
            
            obj.velocity({ 
                "background-color": "#ffcdcd" 
            }, 100, "swing");
            
        } else {
            
            playSound("move"); 
            
            // This is a number tile
            var display = (board[row][col] == 0) ? " " : board[row][col];
            obj.innerText = display;
            
            // Lock the tile
            obj.setAttribute("data-lock", true);
            
            // Do animation
            obj.velocity({
                'background-color': "rgb(238, 238, 238)",
                'font-size': '40px'
            },{
                duration: 100,
                easing: "swing",
                complete: function() {
                    // If this was a 0 or blank then 
                    // reveal the cells next to it
                    if (display == " ") {
                        reveal(row - 1, col - 1);
                        reveal(row - 1, col);
                        reveal(row - 1, col + 1);
                        reveal(row, col - 1);
                        reveal(row, col + 1);
                        reveal(row + 1, col - 1);
                        reveal(row + 1, col);
                        reveal(row + 1, col + 1);
                    }
                },
                loop: false
            });                
        }
    }
    
    // Check for a win
    if (tilesFlipped() == 64) {
        if (intro) intro.stop();
        playSound("win"); 
        showDialogue('<div class="big">You Win!</div><br><span class="yellow">Reload for a rematch!</span>'); 
    }
}

var pause = function() {
    var boardSize   = $('board').getBoundingClientRect();
    var pause = $('pause');

    pause.style.top  = boardSize.top + 'px';
    pause.style.left = boardSize.left + 'px';
    
    if (pause.style.visibility == "hidden") {
        pause.style.visibility = "visible";  
        gameStatus.timerEnabled = false;
        pause.style.opacity = 0;
        pause.velocity({opacity:1},500);
    } else {
        gameStatus.timerEnabled = true;
        pause.style.opacity = 1;
        pause.velocity({opacity:0},{
            duration: 500,
            complete: function(){ pause.style.visibility = "hidden"; }
        });
    }
}

var showDialogue = function(txt, allowClose = true) {
    var board   = $('board');
    var message = $('message');
    
    function show(txt) {
        // Update message with the new values
        message.style.opacity = 0;
        message.style.display = "block";
        $('content').innerHTML = txt;
        
        $('content').onclick = function(){ hideDialogue(); }
        message.onclick = function(){ hideDialogue(); }
        
        if (allowClose == false) {
            $('content').onclick = "";
            message.onclick = "";
        }
        
        // The height and width
        var boardSize   = board.getBoundingClientRect();
        var messageSize = message.getBoundingClientRect();
        
        // Reposition HTML element
        var messageTop  = boardSize.top;
        var messageLeft = (boardSize.width-messageSize.width) / 2;
        messageLeft += 20; // Margin Adjustment
        
        message.style.top  = messageTop + "px";
        message.style.left = messageLeft + 'px';
        
        // Calculate End
        var endPos = (boardSize["height"]-messageSize["height"]) / 2;
        endPos += boardSize.top;
        
        message.velocity({
            top: [endPos + 'px', boardSize.top + 'px'], 
            opacity: 1
        },500);
    }
    
    if (message.style.opacity == 1) {
        if (intro) intro.volume = .1;
        
        message.velocity({
            opacity: 0
        },{
            duration: 100,
            easing: "swing",
            complete: function() {
                show(txt);
            },
            loop: false,
            display: "none"
        });
    } else show(txt);
}

var hideDialogue = function() {
    var message = $('message');
    if (intro) intro.volume = .1;
    
    if (message.style.opacity == 1) {
        message.velocity({
            opacity: 0
        },{
            duration: 100,
            easing: "swing",
            loop: false,
            complete: function() {
                message.style.top = "500px";
            },
            display: "none"
        });
    }
}

var helpToggle = function() {
    var btn = $('btnHelp');
    
    if (btn.innerHTML.includes("Show Help")) {
        btn.innerHTML = "Hide Help";
        
        showDialogue('<div class="yellow big" style="text-decoration: underline">RULES</div><br>' +
            '<ol class="rules"><li>Mark all the bombs</li>' +
            '<li>Flip all the tiles</li>' +
            '<li>Left button reveals a tile</li>' +
            '<li>Right button marks a bomb</li>' +
            '<li>Numbered tiles tell you how many bombs touch that tile</li>' +
            '<li>There are 10 bombs</li></ol><br>' +
            '<div class="yellow">You win when all the tiles are flipped</div>'
        );

    } else {
        btn.innerHTML = "Show Help";
        hideDialogue();
    }
}

var musicToggle = function() {
    var btn = $('btnMusic');
    
    if (btn.innerHTML.includes("Music OFF")) {
        btn.innerHTML = "Music ON";
        intro.stop();
    } else {
        btn.innerHTML = "Music OFF";
        intro = createjs.Sound.play("intro", {loop:-1, delay:0, volume:.1}); 
    }
}

// Start after the page loads
document.addEventListener("DOMContentLoaded", function (event) {
    initSound();
    setInterval(timer, 1000);
    showDialogue('<div class="big">Minesweeper 2018</div><br><span class="yellow">Click to Continue<span>');
    initBoard();
    render();
});
